

function PricesEGM(K,TT,Z,params::AiyagariParametersEGM) # this functions set initial prices assuming no taxes (as a initial guess)
    @unpack β,α,δ,γ,Lbar,χ,ϕ = params
    w = (Z*(1.0-α)*(K/Lbar)^α)^(1/(1.0+ϕ*α))
    L = Lbar*(w^ϕ)
    R = Z*α*(K/L)^(α-1.0) + 1.0 - δ
       
    return AggVarsEGM(R,w,TT)
end

function AiyagariEGM(
    K::T,
    ns::I,
    Transv::Array{T,1},
    states::Array{T,1}, 
    amin::T = 1e-9,
    amax::T = 200.0,
    curv::Real = 4,
    ϕ::T = 0.1,  
    χ::T = 330.0, 
    β::T = 0.98,
    α::T = 0.4,
    δ::T = 0.02,
    γ::T = 1.5,
    TT::T = 0.0,
    Lbar::T = 3.0,
    na::I = 201,
    nd::I = 201) where{T <: Real,I <: Integer}

    #############Params
    Lbar = χ^ϕ   # with s.*y^(1+phi)==1
    params = AiyagariParametersEGM(β,α,δ,γ,Lbar,χ,ϕ)
    AggVars = PricesEGM(K,TT,1.0,params)
    @unpack R,w = AggVars

    ################## Policy grid
    function grid_fun(a_min,a_max,na, pexp)
        x = range(a_min,step=0.5,length=na)
        grid = a_min .+ (a_max-a_min)*(x.^pexp/maximum(x.^pexp))
        return grid
    end
    aGrid = grid_fun(amin,amax,na,curv)
    ################### Distribution grid
    #dGrid = collect(range(aGrid[1],stop = aGrid[end],length = nd))
    #aGrid = collect(range(amin,stop = amax,length = na))
    dGrid=aGrid

   
  #  guess = vcat(10.0 .+ aGrid,10.0 .+ aGrid)
 #   guess = kron(ones(Float64,ns),10 .+ aGrid)
  #  aGridl_t = kron(ones(Float64,ns),aGrid)
  guess = repeat(10 .+ aGrid,ns)
  aGridl_t = repeat(aGrid,ns)    
    

    #return AiyagariModelEGM(params,aGrid,vcat(aGrid,aGrid),na,dGrid,nd,states,ns,Transv),guess,AggVars
    return AiyagariModelEGM(params,aGrid,aGridl_t,na,dGrid,nd,states,ns,Transv),guess,AggVars
end




function interpEGM(pol::AbstractArray,
                grid::AbstractArray,
                x::T,
                na::Integer) where{T <: Real}
    np = searchsortedlast(pol,x)

    # if you give x (current asset) (out of the policy pol), gives ap (saving decision)  out of the grid (grid)
    
    ##Adjust indices if assets fall out of bounds
    (np > 0 && np < na) ? np = np : 
        (np == na) ? np = na-1 : 
            np = 1        
    #@show np
    ap_l,ap_h = pol[np],pol[np+1]
    a_l,a_h = grid[np], grid[np+1] 
    ap = a_l + (a_h-a_l)/(ap_h-ap_l)*(x-ap_l) 
    
    above =  ap > 0.0 
    return above*ap,np
    # return either 0 or ap
end

function get_cEGM(pol::AbstractArray,
               Aggs::AggVarsEGM,
               CurrentAssets::AbstractArray,
               AiyagariModel::AiyagariModelEGM,
               cpol::AbstractArray) 
    
    @unpack params,aGrid,na,ns,states = AiyagariModel
    @unpack  χ,ϕ = params
    X = copy(states)
    X = ((χ*Aggs.w .*states).^(1+ ϕ))/χ
    
    pol = reshape(pol,na,ns)
    for si = 1:ns
        for ai = 1:na
            asi = (si - 1)*na + ai
#            cpol[asi] = Aggs.R*CurrentAssets[asi] + Aggs.w*states[si] - interpEGM(pol[:,si],aGrid,CurrentAssets[asi],na)[1]
             cpol[asi] = Aggs.R*CurrentAssets[asi] + Aggs.TT + X[si] - interpEGM(pol[:,si],aGrid,CurrentAssets[asi],na)[1]
        end
    end
    return cpol
end

function get_ZEGM(pol::AbstractArray,
               Aggs::AggVarsEGM,
               CurrentAssets::AbstractArray,
               AiyagariModel::AiyagariModelEGM,
               cpol::AbstractArray) 
    @unpack params,aGrid,na,ns,states = AiyagariModel
    @unpack  χ, ϕ = params
    X = copy(states)
    X = ((χ*Aggs.w .*states).^(1+ ϕ))/χ 
    Z = copy(cpol)
    pol = reshape(pol,na,ns)
    for si = 1:ns
        for ai = 1:na
            asi = (si - 1)*na + ai
#           cpol[asi] = Aggs.R*CurrentAssets[asi] + Aggs.w*states[si] - interpEGM(pol[:,si],aGrid,CurrentAssets[asi],na)[1]
            Z[asi] = Aggs.R*CurrentAssets[asi] + Aggs.TT + X[si]/(1+ϕ) - interpEGM(pol[:,si],aGrid,CurrentAssets[asi],na)[1] 
        end
    end
    return Z
end


function EulerBackEGM(pol::AbstractArray,
                   Aggs::AggVarsEGM,
                   Aggs_P::AggVarsEGM,
                   AiyagariModel::AiyagariModelEGM,
                   cpol::AbstractArray,
                   apol::AbstractArray)
    
    @unpack params,na,nd,ns,aGridl,Transv,states = AiyagariModel
    @unpack γ,β,χ,ϕ = params

    Trans = reshape(Transv,ns,ns)
  
    
    R_P,w_P = Aggs_P.R,Aggs_P.w
    R,w,TT = Aggs.R,Aggs.w,Aggs.TT
    
   # cp = get_cEGM(pol,Aggs_P,aGridl,AiyagariModel,cpol) #vector ns*na
    Zp  = get_ZEGM(pol,Aggs_P,aGridl,AiyagariModel,cpol) #vector ns*na

    upcp = uPrime(Zp,γ)
    Eupcp = copy(cpol)
    #Eupcp_sp = 0.0
   
    #X = copy(states)
    X = ((χ*Aggs.w .*states).^(1+ ϕ))/χ
    
    for ai = 1:na
        for si = 1:ns
            asi = (si-1)*na + ai
            Eupcp_sp = 0.0
            for spi = 1:ns
                aspi = (spi-1)*na + ai
#               Eupcp_sp += Trans[spi,si]*upcp[aspi]*R_P
                Eupcp_sp += Trans[spi,si]*upcp[aspi] 
            end
            Eupcp[asi] = Eupcp_sp 
        end
    end

    upc = β*Eupcp*R_P    
    c = copy(upc)
    Z = uPrimeInv(upc,γ)

    for ai = 1:na
        for si = 1:ns
            asi = (si-1)*na+ai
#            apol[asi] = (aGridl[asi] + c[asi] - w*states[si])/R
            c[asi]    = Z[asi] + X[si]*ϕ/(1+ϕ)
            apol[asi] = (aGridl[asi] + c[asi] - TT- X[si])/R                  
        end
    end

    return apol,c,upc
end


function SolveEGM(pol::AbstractArray,
                  Aggs::AggVarsEGM,
                  AiyagariModel::AiyagariModelEGM,
                  cpol::AbstractArray,
                  apol::AbstractArray,tol = 1e-10)
    @unpack ns,na = AiyagariModel

    for i = 1:10000
        a = EulerBackEGM(pol,Aggs,Aggs,AiyagariModel,cpol,apol)[1]
        if (i-1) % 50 == 0
            test = abs.(a - pol)/(abs.(a) + abs.(pol))
            #println("iteration: ",i," ",maximum(test))
            if maximum(test) < tol
                println("Solved in ",i," ","iterations")
                break
            end
        end
        pol = copy(a)
    end
    return pol
end

#########################################################
 # Below the functions  for the stationary distribution
#########################################################

function MakeTransMatEGM(pol,AiyagariModel,tmat)
    @unpack ns,na,aGrid,Transv = AiyagariModel
    pol = reshape(pol,na,ns)
    
    Trans = reshape(Transv,ns,ns)
    tmat = copy(tmat)
    tmat  .=0.0

    for a_i = 1:na
        for j = 1:ns
            x,i = interpEGM(pol[:,j],aGrid,aGrid[a_i],na)
            p = (aGrid[a_i] - pol[i,j])/(pol[i+1,j] - pol[i,j])
            p = min(max(p,0.0),1.0)
            sj = (j-1)*na
            for k = 1:ns
                sk = (k-1)*na
                tmat[sk+i+1,sj+a_i] = p * Trans[k,j]
                tmat[sk+i,sj+a_i] = (1.0-p) * Trans[k,j]
            end
        end
    end
   
    return tmat
end

function StationaryDistributionEGM(Mat,AiyagariModel)
    @unpack params,ns,nd,na = AiyagariModel 
    x = zeros(Float64,(na*ns,1))
    λ, x = powm!(Mat, ones(ns*na)/(ns*na), maxiter = 100000,tol = 1e-8)
    x = x/sum(x)
    return x
end


function steady(
    initialpol::AbstractArray,
    AiyagariModel::AiyagariModelEGM,
    R::T,
    w::T,
    TT::T = 0.0,    
    tol = 1e-10,maxn = 50)where{T <: Real}
    @unpack params,aGrid,na,dGrid,nd,ns,na = AiyagariModel
    
    tmat = zeros(eltype(initialpol),(na*ns,na*ns))
    cmat = zeros(eltype(initialpol),na*ns)
    pol = initialpol
    
    Aggs = AggVarsEGM(R,w,TT)
    cmat .= 0.0
    pol = SolveEGM(pol,Aggs,AiyagariModel,cmat,cmat)

    #Stationary transition
    tmat .= 0.0
    trans = MakeTransMatEGM(pol,AiyagariModel,tmat)
    D = StationaryDistributionEGM(trans,AiyagariModel)
    
    #Aggregate savings
    dGridl_t = repeat(dGrid,ns)
    EA = dot(dGridl_t,D)
   
    #println(" Int. rate: ",Aggs.R," Bond S.: ",EA)
    
    cmat .= 0.0
    polA,polC,upc = EulerBackEGM(pol,Aggs,Aggs,AiyagariModel,cmat,cmat)
    
    
    return polA,polC,D,EA,Aggs,trans,upc  
end


